/*
    Copyright (c) 2010, 2011 aizo ag, Zurich, Switzerland

    Author: Sergey 'Jin' Bostandzhyan <jin@dev.digitalstrom.org>

    This file is part of dS Assistant.

    dS-Assistant is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    dS-Assistant is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.

    NOTE: all APIs, especially the JSON API, offered by this application
          are subject to change without further notice.

    Various timeutil functions are borrowed from the MediaTomb project:
    http://mediatomb.cc/
*/

#ifdef HAVE_CONFIG_H
    #include "autoconfig.h"
#endif

#include <sstream>
#include <fstream>
#include <iomanip>
#include <string.h>
#include <stdlib.h>
#include <errno.h>
#include <limits.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <net/if.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include "util.h"
#include "exceptions.h"

namespace dSA
{

static const char *k_hex = "0123456789ABCDEF";

std::string util::str(int i)
{
    std::ostringstream out;
    out << i;
    return out.str();
}

std::string util::str(uint32_t u)
{
    std::ostringstream out;
    out << u;
    return out.str();
}

std::string util::str(unsigned long l)
{
    std::ostringstream out;
    out << l;
    return out.str();
}

std::string util::str(long long int i)
{
    std::ostringstream out;
    out << i;
    return out.str();
}

std::string util::str(time_t t)
{
    std::ostringstream out;
    out << t;
    return out.str();
}

std::string util::str(dsid_t dsid)
{
    std::ostringstream str;
    for (int i = 0; i < 12; ++i)
    {
        str << std::hex
            << std::setw(2)
            << std::setfill('0')
            << (int)dsid.id[i];
      }
      return str.str();
}

std::string util::hex(int i, int width)
{
    std::ostringstream str;
    str << "0x" << std::hex << std::setw(width) << std::setfill('0') << i;
    return str.str();
}

std::string util::hex(uint32_t l, int width)
{
    std::ostringstream str;
    str << "0x" << std::hex << std::setw(width) << std::setfill('0') << l;
    return str.str();
}

std::string util::hex(uint8_t *data, int length)
{
    std::ostringstream str;
    for (int i = 0; i < length; i++)
         str << std::hex << std::setw(2) << std::setfill('0') << (int) data[i];
    return str.str();
}

int util::str2int(std::string str)
{
    if (str.empty())
    {
        throw _Exception("could not convert string to integer, empty string");
    }

    char *endptr;
    errno = 0;
    long int i = strtol(str.c_str(), &endptr, 10);

    if ((errno == ERANGE && (i == LONG_MAX || i == LONG_MIN)) ||
        (errno != 0 && i == 0))
    {
        throw _Exception("could not convert string to integer");
    }

    if (endptr == str.c_str())
    {
        throw _Exception("could not convert string to integer");
    }

    if ((i > INT_MAX) || (i < INT_MIN))
    {
        throw _Exception("could not convert string to integer, "
                         "value out of range");
    }
    return (int)i;
}

long long int util::str2llint(std::string str)
{
    if (str.empty())
    {
        throw _Exception("could not convert string to unsigned long, "
                         "empty string");
    }

    char *endptr;
    errno = 0;
    long long int i = strtoll(str.c_str(), &endptr, 10);

    if ((errno == ERANGE && (i == LLONG_MAX || i == LLONG_MIN)) ||
        (errno != 0 && i == 0))
    {
        throw _Exception("could not convert string to unsigned long");
    }

    if (endptr == str.c_str())
    {
        throw _Exception("could not convert string to unsigned long");
    }

    return i;
}

unsigned long util::str2ulong(std::string str)
{
    long long int i = str2llint(str);
    if (((unsigned long)i > ULONG_MAX) || (i < 0))
    {
        throw _Exception("could not convert string to unsigned long, "
                         "value out of range");
    }
    return (unsigned long)i;
}

double util::str2double(std::string str)
{
    if (str.empty())
    {
        throw _Exception("could not convert string to double, "
                         "empty string");
    }

    double d;

    std::istringstream(str) >> d;
    return d;
}

uint64_t util::str2uint64_t(std::string str)
{
    if (str.empty())
    {
        throw _Exception("could not convert string to uint32_t, empty string");
    }

    uint64_t u;
    std::istringstream(str) >> u;
    return u;
}

uint32_t util::str2uint32_t(std::string str)
{
    if (str.empty())
    {
        throw _Exception("could not convert string to uint32_t, empty string");
    }

    uint32_t u;
    std::istringstream(str) >> u;
    return u;
}

uint16_t util::str2uint16_t(std::string str)
{
    if (str.empty())
    {
        throw _Exception("could not convert string to uint32_t, empty string");
    }

    uint16_t u;
    std::istringstream(str) >> u;
    return u;
}


time_t util::str2time_t(std::string str)
{
    if (str.empty())
    {
        throw _Exception("could not convert string to time_t, "
                         "empty string");
    }

    time_t t;
    std::istringstream(str) >> t;
    return t;
}


dsid_t util::str2dsid(std::string dsid_str)
{
    dsid_t dsid;

    if (dsid_str.length() != 24)
    {
        throw _Exception("can not convert from string, invalid dSID");
    }

    size_t i = 0;
    size_t count = 0;
    while (i < 24)
    {
        std::string number = dsid_str.substr(i, 2);
        dsid.id[count] = (unsigned char)strtol(number.c_str(), NULL, 16);
        i = i + 2;
        count++;
    }

    return dsid;
}

uint64_t util::dsid2upper(const dsid_t& dsid)
{
    uint64_t u = 0;
    for (int i = 0; i < 8; ++i)
    {
        u |= ((uint64_t)dsid.id[i]) << 8*(7-i);
    }
    return u;
}

uint32_t util::dsid2lower(const dsid_t& dsid)
{
    uint32_t l = 0;
    for (int i = 0; i < 4; ++i)
    {
        l |= dsid.id[8+i] << 8*(3-i);
    }

    return l;
}
std::string util::mac2str()
{
    std::string mac;
#ifdef __linux__
    std::string sysfile = "/sys/class/net/eth0/address";
    std::ifstream inFile(sysfile.c_str());
    if(!inFile.is_open())
    {
        mac = "00:00:00:00:00:00";
    }
    else
    {
        inFile >> mac;
    }
#else
#error "Unsupported platform"
#endif
    return mac;
}

void util::strsplit(const std::string& str, char d, std::list<std::string>& list)
{
    const char* s(str.c_str());
    while (*s)
    {
        const char* item(s);
        while (*s && *s != d)
        {
            s++;
        }
        list.push_back(std::string(item, s-item));
        if (*s && !*(++s))
        {
            list.push_back("");
        }
    }
}

std::string webutil::url_unescape(const std::string& url)
{
    const char *data = url.c_str();
    size_t len = url.size();

    std::ostringstream buf;

    size_t i = 0;
    while (i < len)
    {
        char c = data[i++];
        if (c == '%')
        {
            if (i + 2 > len)
            {
                break;
            }
            char chi = data[i++];
            char clo = data[i++];
            int hi, lo;

            const char *pos;

            pos = strchr(k_hex, chi);
            if (!pos)
                hi = 0;
            else
                hi = pos - k_hex;

            pos = ::strchr(k_hex, clo);
            if (!pos)
                lo = 0;
            else
                lo = pos - k_hex;

            int ascii = (hi << 4) | lo;
            buf << (char)ascii;
        }
        else if (c == '+')
        {
            buf << ' ';
        }
        else
        {
            buf << c;
        }
    }
    return buf.str();
}

std::string webutil::url_escape(const std::string& url)
{
    static const char *hex = "0123456789ABCDEF";
    const char *data = url.c_str();
    size_t len = url.length();

    std::ostringstream escaped;
    for (size_t i = 0; i < len; i++)
    {
        unsigned char c = (unsigned char)data[i];
        if ((c >= '0' && c <= '9') || (c >= 'A' && c <= 'Z') ||
            (c >= 'a' && c <= 'z') || c == '_' || c == '-')
        {
            escaped << (char)c;
        }
        else
        {
            int hi = c >> 4;
            int lo = c & 15;
            escaped << '%' << hex[hi] << hex[lo];
        }
    }
    return escaped.str();
}

std::string webutil::command_from_uri(const std::string& uri)
{
    size_t slash = uri.rfind('/');
    if (slash != std::string::npos)
    {
        return uri.substr(slash + 1);
    }
    else
    {
        return uri;
    }
}

void webutil::parse_query(const std::string& query_string, query &parameters)
{
    size_t ampersand;
    std::string pair;
    std::string qstr = std::string(query_string);

    if (query_string.empty())
    {
        return;
    }

    do
    {
        ampersand = qstr.find('&');
        if (ampersand != std::string::npos)
        {
            pair = qstr.substr(0, ampersand);
            qstr = qstr.substr(ampersand + 1);
        }
        else
        {
            pair = qstr;
            qstr = "";
        }

        size_t eq = pair.find('=');
        // skip broken pair
        if (eq == std::string::npos)
        {
            if (ampersand == std::string::npos)
            {
                break;
            }
            else
            {
                continue;
            }
        }

        std::string key = pair.substr(0, eq);
        std::string val = pair.substr(eq + 1);

        if (!key.empty())
        {
            parameters[webutil::url_unescape(key)] = webutil::url_unescape(val);
        }
    } while (ampersand != std::string::npos);
}

std::string webutil::construct_query(const query& q)
{
    query::const_iterator i = q.begin();
    std::ostringstream out;

    while (i != q.end())
    {
        out << url_escape(i->first);
        out << '=';
        out << url_escape(i->second);

        i++;

        if (i != q.end())
        {
            out << '&';
        }
    }

    return out.str();
}

std::string webutil::http_code_msg(const int code)
{
    std::string msg = "Unrecognized Error Code";

    switch (code)
    {
        case 200:
            msg = "OK";
            break;
        case 400:
            msg = "Bad Request";
            break;
        case 403:
            msg = "Forbidden";
            break;
        case 404:
            msg = "Not Found";
            break;
        case 500:
            msg = "Internal Server Error";
            break;
        default:
            break;
    }

    return msg;
}

std::string webutil::render_http_header(const int code,
                                        const std::string& content_type,
                                        const std::string& attachment)
{
    std::ostringstream sstream;
    sstream << "HTTP/1.1 " << code << ' ' << http_code_msg(code) << "\r\n";
    sstream << "Server: " << PACKAGE_NAME "/" PACKAGE_VERSION "\r\n";
    sstream << "Content-Type: " << content_type << "; charset=utf-8" "\r\n";
    if (!attachment.empty())
    {
        sstream << "Content-Disposition: attachment; filename=\"";
        sstream << attachment << "\"" "\r\n";
    }
    sstream << "\r\n";
    return sstream.str();
}

bool fileutil::exists(std::string file, bool need_dir)
{
    int ret;
    struct stat statbuf;

    ret = stat(file.c_str(), &statbuf);
    if (ret != 0) return false;

    if ((need_dir && (!S_ISDIR(statbuf.st_mode))) ||
       (!need_dir && (S_ISDIR(statbuf.st_mode)))) return false;

    return true;
}

size_t fileutil::fsize(std::string file)
{
    int ret;
    struct stat statbuf;

    ret = stat(file.c_str(), &statbuf);
    if (ret != 0) return 0;

    return statbuf.st_size;
}

std::string netutil::if_to_ip(std::string interface)
{
    struct if_nameindex *iflist = NULL;
    struct if_nameindex *iflist_free = NULL;
    struct ifreq if_request;
    struct sockaddr_in local_address;
    int local_socket;

    if (interface.empty())
    {
        throw _Exception("conversion from invalid interface");
    }

    local_socket = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
    if (local_socket < 0)
    {
        throw _Exception(std::string("failed to create local socket: ") +
                         strerror(errno));
    }

    iflist = iflist_free = if_nameindex();
    if (iflist == NULL)
    {
        close(local_socket);
        throw _Exception(std::string("failed to retrieve interface list: ") +
                         strerror(errno));
    }

    while (iflist->if_index || iflist->if_name)
    {
        if (interface == iflist->if_name)
        {
            strncpy(if_request.ifr_name, iflist->if_name, IF_NAMESIZE);
            if (ioctl(local_socket, SIOCGIFADDR, &if_request) != 0)
            {
                close(local_socket);
                if_freenameindex(iflist_free);
                throw _Exception(
                        std::string("failed to determine interface address: ") +
                        strerror(errno));
            }

            memcpy(&local_address, &if_request.ifr_addr,
                   sizeof(if_request.ifr_addr));
            std::string ip = inet_ntoa(local_address.sin_addr);
            if_freenameindex(iflist_free);
            close(local_socket);
            return ip;
        }
        iflist++;
    }

    close(local_socket);
    if_freenameindex(iflist_free);
    throw _Exception("interface not found");
}

std::string timeutil::epoch2local(time_t t)
{
    struct tm result;
    char buf[27];
    localtime_r(&t, &result);
    asctime_r(&result, buf);
    buf[26] = '\0';
    return buf;
}

// the functions below are borrowed from the MediaTomb project
int timeutil::compareTimespecs(struct timespec *a,  struct timespec *b)
{
    if (a->tv_sec < b->tv_sec)
    {
        return 1;
    }
    if (a->tv_sec > b->tv_sec)
    {
        return -1;
    }
    if (a->tv_nsec < b->tv_nsec)
    {
        return 1;
    }
    if (a->tv_nsec > b->tv_nsec)
    {
        return -1;
    }
    return 0;
}

void timeutil::getTimespecNow(struct timespec *ts)
{
    struct timeval tv;
    int ret = gettimeofday(&tv, NULL);
    if (ret != 0)
    {
        throw _Exception("gettimeofday failed: " +
                         std::string(strerror(errno)));
    }

    ts->tv_sec = tv.tv_sec;
    ts->tv_nsec = tv.tv_usec * 1000;
}

void timeutil::getTimespecAfterMillis(long delta, struct timespec *ret,
                                      struct timespec *start)
{
    struct timespec now;
    if (start == NULL)
    {
        getTimespecNow(&now);
        start = &now;
    }
    ret->tv_sec = start->tv_sec + delta / 1000;
    ret->tv_nsec = (start->tv_nsec + (delta % 1000) * 1000000);
    if (ret->tv_nsec >= 1000000000) // >= 1 second
    {
        ret->tv_sec ++;
        ret->tv_nsec -= 1000000000;
    }
}

} // namespace
